@@ -1,6 +1,8 @@
+from __future__ import annotations
+
 import sys
 
-from word_search_generator.cli import main
+from .cli import main
 
 if __name__ == "__main__":
     sys.exit(main())
