@@ -1,10 +1,17 @@
+from __future__ import annotations
+
 import argparse
 import pathlib
 import sys
-from typing import Optional, Sequence
+from datetime import datetime
+from typing import Sequence
 
-from word_search_generator import WordSearch, __app_name__, __version__, config, utils
-from word_search_generator.types import Direction
+from . import WordSearch, __app_name__, __version__, config, utils
+from .mask import shapes
+from .mask.bitmap import Image
+from .word import Direction
+
+BUILTIN_SHAPES = shapes.get_shape_objects()
 
 
 class RandomAction(argparse.Action):
@@ -47,7 +54,7 @@ class SizeAction(argparse.Action):
         setattr(namespace, self.dest, values)
 
 
-def main(argv: Optional[Sequence[str]] = None) -> int:
+def main(argv: Sequence[str] | None = None) -> int:
     """Word Search Generator CLI.
 
     Args:
@@ -67,35 +74,21 @@ Valid Directions: {', '.join([d.name for d in Direction])}
         prog=__app_name__,
         formatter_class=argparse.RawDescriptionHelpFormatter,
     )
-    group = parser.add_mutually_exclusive_group()
-    group.add_argument(
+    words_group = parser.add_mutually_exclusive_group()
+    secret_words_group = parser.add_mutually_exclusive_group()
+    mask_group = parser.add_mutually_exclusive_group()
+    words_group.add_argument(
         "words",
         type=str,
         nargs="*",
         default=sys.stdin,
         help="Words to include in the puzzle (default: stdin).",
     )
-    group.add_argument(
-        "-r",
-        "--random",
-        type=int,
-        action=RandomAction,
-        help="Generate {n} random words to include in the puzzle.",
-    )
-    group2 = parser.add_mutually_exclusive_group()
-    group2.add_argument(
-        "-x",
-        "--secret-words",
-        type=str,
-        default="",
-        help="Secret bonus words not included in the word list.",
-    )
-    group2.add_argument(
-        "-rx",
-        "--random-secret-words",
-        type=int,
-        action=RandomAction,
-        help="Generate {n} random secret words to include in the puzzle.",
+    parser.add_argument(
+        "-c",
+        "--cheat",
+        action="store_true",
+        help="Show the puzzle solution or include it within the `-o, --output` file.",
     )
     # new implementation of -l, --level allowing for more flexibility
     # keeping -l, --level for backwards compatibility
@@ -109,11 +102,53 @@ Valid Directions: {', '.join([d.name for d in Direction])}
             puzzle words can go. See valid arguments above.",
     )
     parser.add_argument(
-        "-xd",
-        "--secret-difficulty",
-        action=DifficultyAction,
-        help="Difficulty level (numeric) or cardinal directions \
-            secret puzzle words can go. See valid arguments above.",
+        "-f",
+        "--format",
+        choices=["CSV", "JSON", "PDF", "csv", "json", "pdf"],
+        metavar="EXPORT_FORMAT",
+        help='Puzzle output format \
+            (choices: "CSV", "JSON", "PDF").',
+    )
+    mask_group.add_argument(
+        "-im",
+        "--image-mask",
+        type=pathlib.Path,
+        help="Mask the puzzle to a provided image \
+            (accepts: BMP, JPEG, PNG).",
+    )
+    mask_group.add_argument(
+        "-m",
+        "--mask",
+        choices=BUILTIN_SHAPES,
+        metavar="MASK_SHAPE",
+        help=f"Mask the puzzle to a shape \
+            (choices: {', '.join(BUILTIN_SHAPES)}).",
+    )
+    parser.add_argument(
+        "-o",
+        "--output",
+        type=pathlib.Path,
+        help="Output path for the saved puzzle.",
+    )
+    parser.add_argument(
+        "-pm",
+        "--preview-masks",
+        action="store_true",
+        help="Preview all built-in mask shapes.",
+    )
+    words_group.add_argument(
+        "-r",
+        "--random",
+        type=int,
+        action=RandomAction,
+        help="Generate {n} random words to include in the puzzle.",
+    )
+    secret_words_group.add_argument(
+        "-rx",
+        "--random-secret-words",
+        type=int,
+        action=RandomAction,
+        help="Generate {n} random secret words to include in the puzzle.",
     )
     parser.add_argument(
         "-s",
@@ -122,28 +157,40 @@ Valid Directions: {', '.join([d.name for d in Direction])}
         type=int,
         help=f"{config.min_puzzle_size} <= puzzle size <= {config.max_puzzle_size}",
     )
-    parser.add_argument(
-        "-c",
-        "--cheat",
-        action="store_true",
-        help="Show the puzzle solution or include it within the `-o, --output` file.",
+    secret_words_group.add_argument(
+        "-x",
+        "--secret-words",
+        type=str,
+        default="",
+        help="Secret bonus words not included in the word list.",
     )
     parser.add_argument(
-        "-o",
-        "--output",
-        type=pathlib.Path,
-        help="Output path for saved puzzle. Specify export type by appending "
-        "'.pdf' or '.csv' to your path (default: PDF).",
+        "-xd",
+        "--secret-difficulty",
+        action=DifficultyAction,
+        help="Difficulty level (numeric) or cardinal directions \
+            secret puzzle words can go. See valid arguments above.",
     )
     parser.add_argument(
         "--version", action="version", version=f"%(prog)s {__version__}"
     )
     args = parser.parse_args(argv)
 
+    # check for mask preview first
+    if args.preview_masks:
+        preview_size = 21
+        for shape in BUILTIN_SHAPES:
+            mask = eval(f"shapes.{shape}")()
+            mask.generate(preview_size)
+            print(f"{shape}")
+            mask.show(True)
+            print()
+        return 0
+
     # process puzzle words
     words = ""
     if args.random:
-        words = utils.get_random_words(args.random)
+        words = ",".join(utils.get_random_words(args.random))
     else:
         if isinstance(args.words, list):
             # needed when words were provided as "command, then, space"
@@ -157,7 +204,7 @@ Valid Directions: {', '.join([d.name for d in Direction])}
     secret_words = (
         args.secret_words
         if args.secret_words
-        else utils.get_random_words(args.random_secret_words)
+        else ",".join(utils.get_random_words(args.random_secret_words))
         if args.random_secret_words
         else ""
     )
@@ -176,9 +223,21 @@ Valid Directions: {', '.join([d.name for d in Direction])}
         secret_level=args.secret_difficulty,
     )
 
+    # apply masking is specified
+    if args.mask:
+        puzzle.apply_mask(eval(f"shapes.{args.mask}")())
+    if args.image_mask:
+        puzzle.apply_mask(Image(args.image_mask))
+
     # show the result
-    if args.output:
-        foutput = puzzle.save(path=args.output, solution=args.cheat)
+    if args.output or args.format:
+        format = args.format if args.format else "PDF"
+        path = (
+            args.output
+            if args.output
+            else f"WordSearchPuzzle {datetime.now()}.{format.lower()}"
+        )
+        foutput = puzzle.save(path=path, format=format, solution=args.cheat)
         print(f"Puzzle saved: {foutput}")
     else:
         puzzle.show(solution=args.cheat)
