@@ -1,21 +1,114 @@
 from __future__ import annotations
 
+import copy
+import math
 import random
 import string
-from typing import TYPE_CHECKING, Any, Iterable, List, Tuple
+import sys
+from math import log2
+from typing import TYPE_CHECKING, Any, Iterable, Sized
 
-from word_search_generator import config
-from word_search_generator.types import (
-    Direction,
-    DirectionSet,
-    Key,
-    Puzzle,
-    Word,
-    Wordlist,
-)
+from . import config
+from .word import Direction, Word
 
 if TYPE_CHECKING:  # pragma: no cover
-    from word_search_generator import WordSearch
+    from . import DirectionSet, Key, Puzzle, WordSearch
+    from .word import Wordlist
+
+
+def calc_puzzle_size(words: Wordlist, level: Sized, size: int | None = None) -> int:
+    """Calculate the puzzle grid size."""
+    all_words = list(word.text for word in words)
+    longest_word_length = len(max(all_words, key=len))
+    shortest_word_length = len(min(all_words, key=len))
+    if not size:
+        longest = max(10, longest_word_length)
+        # calculate multiplier for larger word lists so that most have room to fit
+        multiplier = len(all_words) / 15 if len(all_words) > 15 else 1
+        # level lengths in config.py are nice multiples of 2
+        l_size = log2(len(level)) if level else 1  # protect against log(0) in tests
+        size = round(longest + l_size * 2 * multiplier)
+    else:
+        if size < shortest_word_length:
+            print(
+                "Puzzle sized adjust to fit word with the shortest length.",
+                file=sys.stderr,
+            )
+            size = shortest_word_length + 1
+    return size
+
+
+def build_puzzle(size: int, char: str) -> Puzzle:
+    return [[char] * size for _ in range(size)]
+
+
+def round_half_up(n: float, decimals: int = 0) -> float | Any:  # mypy 0.95+ weirdness
+    """Round numbers in a consistent and familiar format."""
+    multiplier = 10**decimals
+    return math.floor(n * multiplier + 0.5) / multiplier
+
+
+def float_range(
+    start: int | float,
+    stop: int | float | None = None,
+    step: int | float | None = None,
+):
+    """Generate a float-based range for iteration."""
+    start = float(start)
+    if stop is None:
+        stop = start + 0.0
+        start = 0.0
+    if step is None:
+        step = 1.0
+    count = 0
+    while True:
+        temp = float(start + count * step)
+        if step > 0 and temp >= stop:
+            break
+        elif step < 0 and temp <= stop:
+            break
+        yield temp
+        count += 1
+
+
+def distance(x: int, y: int, ratio: float) -> float:
+    """Calculate the distance between two coordinates on a grid."""
+    return math.sqrt(math.pow(y * ratio, 2) + math.pow(x, 2))
+
+
+def in_bounds(x: int, y: int, width: int, height: int) -> bool:
+    """Validate position (x, y) is within the supplied bounds."""
+    return x >= 0 and x < width and y >= 0 and y < height
+
+
+def find_bounding_box(
+    grid: list[list[str]], edge: str = config.ACTIVE
+) -> tuple[tuple[int, int], tuple[int, int]]:
+    """Bounding box of the masked area as a rectangle defined
+    by a Tuple of (top-left edge as x, y, bottom-right edge as x, y)"""
+    size = len(grid)
+    min_y = 0
+    for i, r in enumerate(grid):
+        if edge in r:
+            min_y = i
+            break
+    max_y = size
+    for i, r in enumerate(reversed(grid)):
+        if edge in r:
+            max_y = size - 1 - i
+            break
+    cols = [list(c) for c in zip(*grid)]  # mypy not playing nice w/ `list(zip(*grid))``
+    min_x = 0
+    for i, r in enumerate(cols):
+        if edge in r:
+            min_x = i
+            break
+    max_x = size
+    for i, r in enumerate(reversed(cols)):
+        if edge in r:
+            max_x = size - 1 - i
+            break
+    return ((min_x, min_y), (max_x, max_y))
 
 
 def cleanup_input(words: str, secret: bool = False) -> Wordlist:
@@ -71,7 +164,7 @@ def word_contains_word(words: Wordlist, word: str) -> bool:
 
 
 def validate_direction_iterable(
-    d: Iterable[str | Tuple[int, int] | Direction]
+    d: Iterable[str | tuple[int, int] | Direction]
 ) -> DirectionSet:
     """Validates that all the directions in d are found as keys to
     config.dir_moves and therefore are valid directions."""
@@ -103,6 +196,8 @@ def validate_level(d) -> DirectionSet:
     if isinstance(d, str):  # comma-delimited list
         return validate_direction_iterable(d.split(","))
     if isinstance(d, Iterable):  # probably used by external code
+        if not d:
+            raise ValueError("Empty iterable provided.")
         return validate_direction_iterable(d)
     raise TypeError(f"{type(d)} given, not str, int, or Iterable[str]\n{d}")
 
@@ -111,50 +206,64 @@ def direction_set_repr(ds: DirectionSet) -> str:
     return ("'" + ",".join(d.name for d in ds) + "'") if ds else "None"
 
 
-def highlight_solution(puzzle: WordSearch) -> Puzzle:
+def highlight_solution(ws: WordSearch) -> Puzzle:
     """Add highlighting to puzzle solution."""
-    output: Puzzle = []
-    for r, line in enumerate(puzzle.puzzle):
-        line_chars = []
-        # check to see if character if part of the solution
-        for c, char in enumerate(line):
-            if puzzle.solution[r][c]:
-                line_chars.append(f"\u001b[1m\u001b[31m{char}\u001b[0m")
-            else:
-                line_chars.append(f"{char}")
-        output.append(line_chars)
+    output: Puzzle = copy.deepcopy(ws.puzzle)
+    for word in ws.placed_words:
+        if (
+            word.start_column is None
+            or word.start_row is None
+            or word.direction is None
+        ):  # only here for mypy
+            continue  # pragma: no cover
+        x = word.start_column
+        y = word.start_row
+        for char in word.text:
+            output[y][x] = f"\u001b[1m\u001b[31m{char}\u001b[0m"
+            x += word.direction.c_move
+            y += word.direction.r_move
     return output
 
 
-def make_header(puzzle: Puzzle, text: str) -> str:
-    """Generate a header that fits the current puzzle."""
-    hr = "-" * max(11, (len(puzzle) * 2 - 1))
+def make_header(size: int, text: str) -> str:
+    """Generate a header that fits the current puzzle size."""
+    hr = "-" * max(11, (size * 2 - 1))
     padding = " " * ((len(hr) - len(text)) // 2)
     return f"""{hr}
 {padding}{text}{padding}
 {hr}"""
 
 
-def stringify(puzzle: Puzzle) -> str:
+def stringify(puzzle: Puzzle, bbox: tuple[tuple[int, int], tuple[int, int]]) -> str:
     """Convert puzzle array of nested lists into a string."""
+    min_x, min_y = bbox[0]
+    max_x, max_y = bbox[1]
     output = []
-    for line in puzzle:
-        output.append(" ".join(line))
+    offset = " " if max_x - min_x < 5 else ""
+    for line in puzzle[min_y : max_y + 1]:
+        output.append(
+            offset + " ".join([c if c else " " for c in line[min_x : max_x + 1]])
+        )
     return "\n".join(output)
 
 
-def format_puzzle_for_show(puzzle: WordSearch, show_solution: bool = False) -> str:
-    header = make_header(puzzle.puzzle, "WORD SEARCH")
-    word_list = get_word_list_str(puzzle.key)
+def format_puzzle_for_show(ws: WordSearch, show_solution: bool = False) -> str:
+    word_list = get_word_list_str(ws.key)
     # highlight solution if provided
-    puzzle_list = highlight_solution(puzzle) if show_solution else puzzle.puzzle
+    puzzle_list = highlight_solution(ws) if show_solution else ws.puzzle
+    # calculate header length based on cropped puzzle size to account for masks
+    header_width = ws.bounding_box[1][0] - ws.bounding_box[0][0] + 1
+    header = make_header(header_width, "WORD SEARCH")
+    answer_key_intro = (
+        "Answer Key (*Secret Words)" if ws.placed_secret_words else "Answer Key"
+    )
     return f"""{header}
-{stringify(puzzle_list)}
+{stringify(puzzle_list, ws.bounding_box)}
 
 Find these words: {word_list if word_list else '<ALL SECRET WORDS>'}
-* Words can go {get_level_dirs_str(puzzle.level)}.
+* Words can go {get_level_dirs_str(ws.level)}.
 
-Answer Key: {get_answer_key_str(puzzle.placed_words)}"""
+{answer_key_intro}: {get_answer_key_str(ws.placed_words, ws.bounding_box)}"""
 
 
 def get_level_dirs_str(level: DirectionSet) -> str:
@@ -169,27 +278,42 @@ def get_word_list_str(key: Key) -> str:
     return ", ".join(get_word_list_list(key))
 
 
-def get_word_list_list(key: Key) -> List[str]:
+def get_word_list_list(key: Key) -> list[str]:
     """Return all placed puzzle words as a list (excluding secret words)."""
     return [k for k in sorted(key.keys()) if not key[k]["secret"]]
 
 
-def get_answer_key_list(words: Wordlist) -> List[Any]:
-    """Return a easy to read answer key for display/export."""
-    keys = []
-    for w in sorted(words, key=lambda word: word.text):
-        keys.append(w.key_string)
-    return keys
+def get_answer_key_list(
+    words: Wordlist, bbox: tuple[tuple[int, int], tuple[int, int]]
+) -> list[Any]:
+    """Return a easy to read answer key for display/export. Resulting coordinates
+    will be offset by the supplied values. Used for masked puzzles.
+
+    Args:
+        words (Wordlist): A list of `Word` objects.
+        bbox (Tuple[int, int, int, int]): Puzzle mask bounding box
+        coordinates should be offset by.
+    """
+    return [w.key_string(bbox) for w in sorted(words, key=lambda word: word.text)]
 
 
-def get_answer_key_str(words: Wordlist) -> str:
-    """Return a easy to read answer key for display."""
-    keys = get_answer_key_list(words)
-    return ", ".join(keys)
+def get_answer_key_str(
+    words: Wordlist, bbox: tuple[tuple[int, int], tuple[int, int]]
+) -> str:
+    """Return a easy to read answer key for display. Resulting coordinates
+    will be offset by the supplied values. Used for masked puzzles.
+
+    Args:
+        words (Wordlist): A list of `Word` objects.
+        bbox (Tuple[int, int, int, int]): Puzzle mask bounding box
+        coordinates should be offset by.
+    """
+    return ", ".join(get_answer_key_list(words, bbox))
 
 
-def get_random_words(n: int) -> str:
-    return ",".join(random.sample(WORD_LIST, n))
+def get_random_words(n: int) -> list[str]:
+    """Return a list of random dictionary words."""
+    return random.sample(WORD_LIST, n)
 
 
 WORD_LIST = [
